# -*- coding: utf-8 -*-

import re
import random
import socket
import logging
import xmlrpclib
import simplejson

from datetime import datetime, timedelta

from twootr import web, render, LIMIT, XMLRPC_SERVER, COOKIE_LIFE

class Core(object):
	"""
	Parent object for each pages of Twootr.
	"""
	
	reply = re.compile("@([0-9a-zA-Z]+)")
	link = re.compile("(http://[^ ,322)\b]+)")
	users_color = {}
	
	def __init__(self):
		self.__user = None
	
	def base(self, tpl):
		"""
		Print the final page from the base and the given template.
		"""
		print render.base(tpl, self.get_logged_user())
	
	@staticmethod
	def json(data):
		"""
		Prints the data as a JSON object
		"""
		#web.header("Content-type", "application/json")
		web.header("Content-type", "text/plain; charset=utf-8")
		print simplejson.dumps(data)
	
	def get_logged_user(self):
		"""
		Returns the user from the cookie informations.
		"""
		if not self.__user: # lazy
			session = web.cookies()
			if "nick" in session and len(session.nick):
				self.__user = self.get_user(session.nick)
				return self.__user
			else:
				return None
		else:
			return self.__user
	
	@classmethod
	def get_user(cls, nick=None, jid=None):
		"""
		Gets you the user from it's nick.
		"""
		
		where = {}
		if nick:
			where["nick"] = "'"+nick+"'"
		if jid:
			where["jid"] = "'"+jid+"'"
		
		users = web.select(
			'users',
			where=sqlwhere(where),
			limit=1)
		
		users = list(users)
		if users:
			return users[0]
		else:
			return None
	
	@classmethod
	def add_user(cls, nick, jid):
		"""
		Adds the given user to Twootr.
		"""
		web.insert('users', nick=nick, jid=jid)
	
	@classmethod
	def add_friendship(cls, follower, followee):
		"""
		Add a friendship relation between those two guys.
		"""
		web.insert('friendships', from_id=follower.id, to_id=followee.id)
	
	@classmethod
	def del_friendship(cls, follower, followee):
		"""
		Delete the friendship between those two guys.
		"""
		
		web.delete('friendships', where=sqlwhere(
			{"from_id" : follower.id,
			 "to_id" : followee.id }))
	
	@classmethod
	def get_friends(cls, user, full=False):
		"""
		Return the friends of the given user
		"""
		if not full:
			friends = web.select(
				'friendships',
				where=sqlwhere(
					{'friendships.from_id':user.id}))
		else:
			friends = web.select(
				['friendships', 'users'],
				where=sqlwhere(
					{'friendships.from_id':user.id,
					'friendships.to_id':'users.id'}))
		
		return list(friends)
		
	@classmethod
	def get_followers(cls, user, full=False):
		"""
		Return the followers of the given user
		"""
		if not full:
			followers = web.select(
				'friendships',
				where=sqlwhere(
					{'friendships.to_id':user.id}))
		else:
			followers = web.select(
				['friendships', 'users'],
				where=sqlwhere(
					{'friendships.to_id':user.id,
					'friendships.from_id':'users.id'}))
		
		return list(followers)
	
	@classmethod
	def is_friend_with(cls, user_a, user_b):
		"""
		Tell you if the user A is a friend with the user B
		"""
		friendship = web.select('friendships', where=sqlwhere(
			{"from_id" : user_a.id,
			 "to_id" : user_b.id}),
			limit=1)
		
		return len(list(friendship)) == 1
	
	@classmethod
	def get_last_twoot(cls, user):
		"""
		Returns the last twoot of the given user
		"""
		twoots = web.select(
			['twoots'],
			where=sqlwhere({
				'user_id':user.id}),
			order="id DESC",
			limit=1)
		
		twoots = list(twoots)
		if len(twoots):
			return twoots[0]
		else:
			return None
	
	@classmethod
	def get_twoot_by_id(cls, twoot_id, user=None):
		# criteria
		where = {
			'twoots.user_id':'users.id',
			'twoots.id':twoot_id}
		# mandatory one
		if user:
			where['users.id'] = user.id
		
		twoots = web.select(
			['twoots', 'users'],
			what='twoots.*, users.nick, users.id as user_id',
			where=sqlwhere(where),
			limit=1)
		
		twoots = list(twoots)
		if len(twoots):
			return twoots[0]
		else:
			return None
	
	@classmethod
	def get_twoots(cls, user=None, with_friends=True, last_id=None):
		"""
		Returns the twoots of the given user with its friends' ones.
		"""
		
		# Custom SQL
		friends_id = ""
		if user and with_friends:
			friends = cls.get_friends(user)
			user.to_id = user.id
			friends.append(user)
			# create the friends id list.
			friends_id = " AND %s" % \
				sqlors('users.id = ', [(friend.to_id) for friend in friends])
		
		# Custom SQL
		twoots_id = ""
		if last_id:
			twoots_id = " AND twoots.id > '%d'" % last_id
		
		twoots = web.select(
			['users', 'twoots'],
			what='twoots.*, users.nick, users.id as user_id',
			where=sqlwhere({
				'twoots.user_id':'users.id'}) +
				twoots_id +
				friends_id,
			order="twoots.id DESC",
			limit=LIMIT)
		
		return list(twoots)
	
	@classmethod
	def post_twoot(cls, user, status):
		"""
		Post the twoot
		"""
		# Catch the reply info and get the twoot id that correspond
		twoot = None
		reply_to = cls.reply.search(status)
		
		if reply_to:
			to = cls.get_user(reply_to.group(1))
			if to:
				twoot = cls.get_last_twoot(to)
		
		if twoot:
			web.insert('twoots', user_id=user.id, text=status, reply_to=twoot.id)
		else:
			web.insert('twoots', user_id=user.id, text=status)
	
	@classmethod
	def send_notifications(cls, user, status):
		followers = cls.get_followers(user, full=True)
		followers_jid = [(follower.jid) for follower in followers]
		
		web.debug("Followers:")
		web.debug(followers_jid)
		
		try:
			xmlrpc = xmlrpclib.ServerProxy(XMLRPC_SERVER)
			xmlrpc.broadcast(user.jid, u"%s: %s" % (user.nick, status), followers_jid)
		except socket.error, (msg, num):
			logging.error("send_notifications: %s (%s)" % (num, msg))
	
	@classmethod
	def send_adduser(cls, jid):
		try:
			xmlrpc = xmlrpclib.ServerProxy(XMLRPC_SERVER)
			xmlrpc.adduser(jid)
		except socket.error, (msg, num):
			logging.error("send_notifications: %s (%s)" % (num, msg))
	
	@classmethod
	def logged(cls, nick):
		"""
		Action performed went the user is logged.
		"""
		web.setcookie(
			"nick",
			nick,
			web.httpdate(datetime.now() + timedelta(COOKIE_LIFE)))
		
		return cls.go_home()
	
	@classmethod
	def logout(cls):
		web.setcookie(
			"nick",
			"",
			web.httpdate(datetime.now() + timedelta(-1)))
		
		# not a failure this time ;-)
		return cls.epic_fail()
	
	@staticmethod
	def redirect(url):
		return web.redirect(url, status='302 Redirect')
	
	@staticmethod
	def go_home():
		return web.redirect("/home/", status='302 Redirect')
	
	@staticmethod
	def epic_fail(exception=None):
		"""
		What to do when you don't know what to do.
		"""
		# TODO: log this and have an error page.
		if(exception):
			web.debug(exception)
		return web.redirect("/", status='307 Temporary Redirect')
	
	@classmethod
	def get_user_color(cls, user_id):
		if not user_id in cls.users_color:
			random.seed(user_id)
			cls.users_color[user_id] = "%02x%02x%02x" % (random.randint(0,255),random.randint(0,255),random.randint(0,255))
		
		return cls.users_color[user_id]
	
	@classmethod
	def decorate_twoot(cls, twoot):
		# Time
		twoot.created = twoot.created_at.replace(" ", "T")
		twoot.created += "+00:00"
		
		date, time = twoot.created_at.split(" ")
		year, month, day = date.split("-")
		hour, minute, second = time.split(":")
		dt = datetime(int(year), int(month), int(day), int(hour), int(minute), int(second), )
		
		twoot.created_human = dt.strftime("%a, %d %b %Y %H:%M:%S UTC")
		
		twoot.text = twoot.text.replace("<", "&lt;")
		twoot.text = twoot.text.replace(">", "&gt;")
		twoot.text = twoot.text.replace("&", "&amp;")
		
		# Url
		links = cls.link.findall(twoot.text)
		if links:
			for link in links:
				twoot.text = twoot.text.replace(link, '<a href="%s" rel="no-follow">%s</a>' % (link, link)) 
		
		# Replies
		replies = cls.reply.findall(twoot.text)
		if replies:
			for reply in replies: 
				twoot.text = twoot.text.replace(reply, '<a href="/%s/">%s</a>' % (reply, reply))
		
		# Reply to
		if twoot.reply_to is not None:
			twoot.reply = cls.get_twoot_by_id(twoot.reply_to)
		
		# Color
		twoot.color = cls.get_user_color(twoot.user_id)
		
		return twoot
	
	@classmethod
	def decorate_twoots(cls, twoots):
		"""
		Add the info created and relative_time to the given twoots
		"""
		for twoot in twoots:
			cls.decorate_twoot(twoot)
		
		return twoots
	
	# Done Client Side (thank to date.js)
	"""
	@staticmethod
	def get_relative_time(date_time):
		date, time = date_time.split(" ")
		year, month, day = date.split("-")
		hour, minute, second = time.split(":")
		
		dt = datetime(int(year), int(month), int(day), int(hour), int(minute), int(second))
		now = datetime.now()
		
		delta = now - dt
		if delta < timedelta(0,60):
			return "seconds"
		elif delta < timedelta(0,120):
			return "one minute"
		elif delta < timedelta(0,3600):
			return "%d minutes" % (delta.seconds / 60)
		elif delta < timedelta(1):
			return "%d hours" % (delta.seconds / 3600)
		elif delta < timedelta(2):
			return "one day and %d hours" % (delta.seconds / 3600)
		elif delta < timedelta(7):
			return "%d days" % delta.days
		elif delta < timedelta(31):
			return "%d weeks" % (delta.days / 7)
		elif delta < timedelta(60):
			return "one month and %d days" % ((delta.days - 30) / 7)
		elif delta < timedelta(365):
			return "%d months" % (delta.days / 30)
		elif delta < timedelta(395):
			return "one year and one month"
		elif delta < timedelta(730):
			return "one year and %d months" % ((delta.days - 365) / 30)
		else:
			return "%d years" % (delta.days / 365.25)
	"""

# Monkey Patching for web.py
def sqlwhere(dict, grouping=" AND "):
	"""
	Apparently missing from web.py
	"""
	return grouping.join([("%s = %s" % (k,v)) for k, v in dict.iteritems()])

def sqlors(pattern, list, grouping=" OR "):
	"""
	Apparently buggy from web.py
	"""
	return "(%s)" % grouping.join([("%s%s" % (pattern, v)) for v in list])
